# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import audioop
import time
from collections import deque
from dataclasses import dataclass, field
from enum import Enum
from threading import Event
from typing import Callable, Deque, Optional

from ovos_config import Configuration
from ovos_plugin_manager.stt import StreamingSTT
from ovos_plugin_manager.vad import VADEngine
from ovos_utils.log import LOG
from ovos_bus_client.session import SessionManager
from ovos_dinkum_listener.transformers import AudioTransformersService
from ovos_dinkum_listener.voice_loop.hotwords import HotwordContainer, HotwordState, HotWordException
from ovos_plugin_manager.templates.microphone import Microphone

from ovos_dinkum_listener.plugins import FakeStreamingSTT


class ListeningState(str, Enum):
    DETECT_WAKEWORD = "wakeword"
    WAITING_CMD = "continuous"

    RECORDING = "recording"

    SLEEPING = "sleeping"
    CHECK_WAKE_UP = "wake_up"

    CONFIRMATION = "confirmation"

    BEFORE_COMMAND = "before_cmd"
    IN_COMMAND = "in_cmd"
    AFTER_COMMAND = "after_cmd"


class ListeningMode(str, Enum):
    """ global listening mode """
    WAKEWORD = "wakeword"
    CONTINUOUS = "continuous"
    HYBRID = "hybrid"
    SLEEPING = "sleeping"


@dataclass
class VoiceLoop:
    mic: Microphone
    hotwords: HotwordContainer
    stt: StreamingSTT
    fallback_stt: StreamingSTT
    vad: VADEngine
    transformers: AudioTransformersService

    def start(self):
        raise NotImplementedError()

    def run(self):
        raise NotImplementedError()

    def stop(self):
        raise NotImplementedError()

    @staticmethod
    def debiased_energy(audio_data: bytes, sample_width: int) -> float:
        """Compute RMS of debiased audio."""
        # Thanks to the speech_recognition library!
        # https://github.com/Uberi/speech_recognition/blob/master/speech_recognition/__init__.py
        energy = -audioop.rms(audio_data, sample_width)
        energy_bytes = bytes([energy & 0xFF, (energy >> 8) & 0xFF])
        debiased_energy = audioop.rms(
            audioop.add(audio_data,
                        energy_bytes * (len(audio_data) // sample_width),
                        sample_width),
            sample_width,
        )

        return debiased_energy


@dataclass
class ChunkInfo:
    is_speech: bool = False
    is_listen_sound: bool = False
    energy: float = 0.0


RecordCallback = Callable[[], None]
TextCallback = Callable[[str, dict], None]
AudioCallback = Callable[[bytes, dict], None]
ChunkCallback = Callable[[ChunkInfo], None]


@dataclass
class DinkumVoiceLoop(VoiceLoop):
    speech_seconds: float = 0.3
    silence_seconds: float = 0.7
    timeout_seconds: float = 10.0
    timeout_seconds_with_silence: float = 5.0
    confirmation_seconds: float = 0.5
    num_stt_rewind_chunks: int = 2
    num_hotword_keep_chunks: int = 15
    remove_silence: bool = False
    instant_listen: bool = False
    skip_next_wake: bool = False
    hotword_chunks: Deque = field(default_factory=deque)
    stt_chunks: Deque = field(default_factory=deque)
    stt_audio_bytes: bytes = bytes()
    min_stt_confidence: float = 0.6
    max_transcripts: int = 1
    last_ww: float = -1.0
    speech_seconds_left: float = 0.0
    silence_seconds_left: float = 0.0
    confirmation_seconds_left: float = 0.0
    timeout_seconds_left: float = 0.0
    timeout_seconds_with_silence_left: float = 0.0
    recording_seconds_with_silence_left: float = 0.0
    recording_mode_max_silence_seconds: float = 30.0
    state: ListeningState = ListeningState.DETECT_WAKEWORD
    listen_mode: ListeningMode = ListeningMode.WAKEWORD
    wake_callback: Optional[RecordCallback] = None
    text_callback: Optional[TextCallback] = None
    wakeup_callback: Optional[RecordCallback] = None
    listenword_audio_callback: Optional[AudioCallback] = None
    hotword_audio_callback: Optional[AudioCallback] = None
    stopword_audio_callback: Optional[AudioCallback] = None
    wakeupword_audio_callback: Optional[AudioCallback] = None
    stt_audio_callback: Optional[AudioCallback] = None
    recording_audio_callback: Optional[AudioCallback] = None
    record_end_callback: Optional[RecordCallback] = None
    chunk_callback: Optional[ChunkCallback] = None
    recording_filename: str = "rec"
    is_muted: bool = False
    _is_running: bool = False
    _chunk_info: ChunkInfo = field(default_factory=ChunkInfo)

    @property
    def running(self) -> bool:
        """
        Return true while the loop is running
        """
        return self._is_running is True
    
    def reset_speech_timer(self):
        self.speech_seconds_left = self.speech_seconds
        self.timeout_seconds_left = self.timeout_seconds
        self.timeout_seconds_with_silence_left = self.timeout_seconds_with_silence  

    def start(self):
        """
        Start the Voice Loop; sets the listening mode based on configuration and
        prepares the loop to be run.
        """

        self._is_running = True
        self.state = ListeningState.DETECT_WAKEWORD
        self.last_ww = -1
        listener_config = Configuration().get("listener", {})
        if listener_config.get("continuous_listen", False):
            self.listen_mode = ListeningMode.CONTINUOUS
        elif listener_config.get("hybrid_listen", False):
            self.listen_mode = ListeningMode.HYBRID
        else:
            self.listen_mode = ListeningMode.WAKEWORD

        LOG.info(f"Listening mode: {self.listen_mode}")
        LOG.debug(f"STATE: {self.state}")

    def run(self):
        """
        Run the VoiceLoop so long as `self._is_running` is True
        """
        # Voice command state
        self.speech_seconds_left = self.speech_seconds
        self.silence_seconds_left = self.silence_seconds
        self.timeout_seconds_left = self.timeout_seconds
        self.timeout_seconds_with_silence_left = self.timeout_seconds_with_silence        
        self.state = ListeningState.DETECT_WAKEWORD

        # Keep hotword/STT audio so they can (optionally) be saved to disk
        self.hotword_chunks = deque(maxlen=self.num_hotword_keep_chunks)
        self.stt_audio_bytes = bytes()

        # Audio from just before the wake word is detected is kept for STT.
        # This allows you to speak a command immediately after the wake word.
        n = self.num_stt_rewind_chunks + 1
        if self.listen_mode == ListeningMode.CONTINUOUS:
            self.stt_chunks: Deque[bytes] = deque(maxlen=3 * n)
        else:
            self.stt_chunks: Deque[bytes] = deque(maxlen=n)

        LOG.info(f"Starting loop in mode: {self.listen_mode}")

        while self._is_running:
            # If no audio is provided, raise an exception and stop the loop
            chunk = self.mic.read_chunk()
            if not self._is_running:  # handle shutdown in middle of read_chunk
                break
            assert chunk is not None, "No audio from microphone"

            if self.is_muted:
                # Soft mute
                chunk = bytes(self.mic.chunk_size)

            self._chunk_info.is_speech = False
            self._chunk_info.energy = 0.0

            # State machine:
            #
            # DETECT_HOTWORD -> BEFORE_COMMAND
            # BEFORE_COMMAND -> {IN_COMMAND, AFTER_COMMAND}
            # IN_COMMAND -> AFTER_COMMAND
            # AFTER_COMMAND -> DETECT_HOTWORD
            #

            if self.state == ListeningState.DETECT_WAKEWORD:
                try:
                    if self.listen_mode == ListeningMode.CONTINUOUS:
                        LOG.info(f"Continuous listening mode, updating state")
                        self.state = ListeningState.WAITING_CMD
                        LOG.debug(f"STATE: {self.state}")
                    elif self._detect_ww(chunk):
                        LOG.info("Wakeword detected")
                    elif self._detect_hot(chunk):
                        LOG.info("Hotword detected")
                    else:
                        self.transformers.feed_audio(chunk)
                except HotWordException as e:
                    if self.hotwords.reload_on_failure:
                        LOG.warning(e)
                        self.hotwords.load_hotword_engines()
                    else:
                        raise e

            if self.state == ListeningState.WAITING_CMD:
                self._wait_cmd(chunk)

            elif self.state == ListeningState.RECORDING:
                self._in_recording(chunk)

            elif self.state == ListeningState.SLEEPING:
                self._before_wakeup(chunk)
            elif self.state == ListeningState.CHECK_WAKE_UP:
                self._detect_wakeup(chunk)

            elif self.state == ListeningState.CONFIRMATION:
                LOG.debug("playing listen sound")
                self._confirmation_sound(chunk)

            elif self.state == ListeningState.BEFORE_COMMAND:
                LOG.debug("waiting for speech")
                self._before_cmd(chunk)
            elif self.state == ListeningState.IN_COMMAND:
                LOG.debug("recording speech")
                self._in_cmd(chunk)
            elif self.state == ListeningState.AFTER_COMMAND:
                LOG.info("speech finished")
                self._after_cmd(chunk)

            if self.chunk_callback is not None:
                self._chunk_info.energy = \
                    self.debiased_energy(chunk, self.mic.sample_width)
                self.chunk_callback(self._chunk_info)
        LOG.info(f"Loop stopped running")

    def reset_state(self):
        """
        Reset the internal state to the default
        Continuous Listening -> Waiting for Command
        Wakeword Listening -> Waiting for WW
        Hybrid Listening -> Waiting for WW
        """
        self.recording_seconds_with_silence_left = 0
        if self.listen_mode == ListeningMode.CONTINUOUS:
            self.state = ListeningState.WAITING_CMD
            self.hotwords.state = HotwordState.HOTWORD
        else:
            self.state = ListeningState.DETECT_WAKEWORD
            self.hotwords.state = HotwordState.LISTEN
        LOG.debug(f"STATE: {self.state}")
        LOG.debug(f"HOTWORDS STATE: {self.hotwords.state}")

    def go_to_sleep(self):
        """
        Set the Listening State to "Sleeping" until woken
        """
        self.state = ListeningState.SLEEPING
        LOG.info("entering SLEEP mode, STT will not be executed")
        LOG.debug(f"STATE: {self.state}")

    def wakeup(self):
        """
        Reset the Listening State from "Sleeping" to the default for the current
        Listening Mode.
        """
        self.reset_state()
        LOG.info("wakeup")
        if self.wakeup_callback:
            self.wakeup_callback()

    def start_recording(self, filename: Optional[str] = None):
        """
        Set the listening state to RECORDING and specify a file to record to
        @param filename: filename to record mic input to
        """
        self.recording_seconds_with_silence_left = self.recording_mode_max_silence_seconds
        self.recording_filename = filename or str(time.time())
        LOG.debug(f"Recording to {self.recording_filename}")
        self.state = ListeningState.RECORDING
        LOG.debug(f"STATE: {self.state}")
        if self.wake_callback is not None:
            # emit record_begin
            self.wake_callback()

    def stop_recording(self):
        """
        Stop recording, pass audio and metadata (recording filename) to the
        `recording_audio_callback` method and reset the Listening State
        """
        #  finished recording
        if self.recording_audio_callback is not None:
            metadata = {"recording_name": self.recording_filename}
            self.recording_audio_callback(self.stt_audio_bytes, metadata)
        if self.record_end_callback is not None:
            # emit record_end
            self.record_end_callback()
        LOG.debug("Finished recording")
        self.reset_state()

    def _in_recording(self, chunk: bytes):
        """
        Handle a chunk of audio while in the `RECORDING` state.
        Check for stop words in all cases and pass audio to any loaded audio
        transformers.

        If a "stop" hotword is detected, the appropriate method is called.

        If no "stop" hotword is detected, audio is evaluated by VAD and all
        audio frames are passed to audio transformers.

        @param chunk: bytes of audio captured
        """
        self.hotwords.state = HotwordState.RECORDING
        LOG.debug(f"HOTWORDS STATE: {self.hotwords.state}")
        self.hotwords.update(chunk)
        ww = self.hotwords.found()
        if ww:
            # stop recording
            self.stop_recording()

            self.transformers.feed_hotword(chunk)

            # Callback to handle recorded hotword audio
            if self.stopword_audio_callback is not None:
                hotword_audio_bytes = bytes()
                while self.hotword_chunks:
                    hotword_audio_bytes += self.hotword_chunks.popleft()
                self.stopword_audio_callback(hotword_audio_bytes,
                                             self.hotwords.get_ww(ww))
        else:
            # Recording audio until user requests stop
            self._chunk_info.is_speech = not self.vad.is_silence(chunk)
            self.stt_audio_bytes += chunk
            self.stt_chunks.append(chunk)

            self.transformers.feed_speech(chunk)

            if self._chunk_info.is_speech:
                self.recording_seconds_with_silence_left = self.recording_mode_max_silence_seconds
            # check if maximum silence has been detected
            elif self.recording_seconds_with_silence_left <= 0:
                LOG.info("Recording mode timed out, reached max silence time")
                self.stop_recording()
            else:
                n_chunks = len(chunk) / self.mic.chunk_size
                self.recording_seconds_with_silence_left -= n_chunks * self.mic.seconds_per_chunk

    def _before_wakeup(self, chunk: bytes):
        """
        Handle a chunk of audio as unknown input while sleeping,
        passing to wakeup word detection.
        @param chunk: bytes of audio captured
        """
        self.hotwords.state = HotwordState.LISTEN
        if self._detect_ww(chunk):
            if self.state != ListeningState.CHECK_WAKE_UP:
                LOG.debug(f"STATE: {self.state}")
                self.state = ListeningState.CHECK_WAKE_UP

    def _detect_wakeup(self, chunk: bytes) -> bool:
        """
        Handle a chunk of audio where a hotword has been detected.
        Determines if the detected hotword should exit sleeping mode.
        If a wake word has been spoken multiple times,
        @param chunk: bytes of audio captured
        @return: True if wake up word was detected
        """
        if self.hotwords.state != HotwordState.WAKEUP:
            self.hotwords.state = HotwordState.WAKEUP
            LOG.debug(f"HOTWORDS STATE: {self.hotwords.state}")

        self.hotwords.update(chunk)
        ww = self.hotwords.found()
        if ww:
            LOG.info("waking listener from SLEEP mode")
            # get out of sleep mode
            self.state = self.state.DETECT_WAKEWORD
            self.hotwords.state = HotwordState.LISTEN
            LOG.debug(f"STATE: {self.state}")
            LOG.debug(f"HOTWORDS STATE: {self.hotwords.state}")

            if self.wakeup_callback:
                self.wakeup_callback()

            # Callback to handle recorded hotword audio
            if self.wakeupword_audio_callback is not None:
                hotword_audio_bytes = bytes()
                while self.hotword_chunks:
                    hotword_audio_bytes += self.hotword_chunks.popleft()
                self.wakeupword_audio_callback(hotword_audio_bytes,
                                               self.hotwords.get_ww(ww))

            self.transformers.feed_hotword(chunk)

            if self.record_end_callback is not None:
                # emit record_end
                self.record_end_callback()
            return True
        elif time.time() - self.last_ww > 10:
            # require wake word again
            self.hotwords.state = HotwordState.LISTEN
            self.state = ListeningState.SLEEPING
            LOG.debug(f"STATE: {self.state}")
            LOG.debug(f"HOTWORDS STATE: {self.hotwords.state}")
        return False

    def _detect_hot(self, chunk: bytes) -> bool:
        """
        Check for a hotword in a chunk of unknown audio. If a hotword is
        detected, call `hotword_audio_callback` and pass audio to transformers.
        @param chunk: bytes of audio captured
        @return: True if a hotword was detected
        """
        self.hotwords.state = HotwordState.HOTWORD

        self.hotwords.update(chunk)
        ww = self.hotwords.found()
        if ww:
            # Callback to handle recorded hotword audio
            if self.hotword_audio_callback is not None:
                hotword_audio_bytes = bytes()
                while self.hotword_chunks:
                    hotword_audio_bytes += self.hotword_chunks.popleft()
                self.hotword_audio_callback(hotword_audio_bytes,
                                            self.hotwords.get_ww(ww))
                self.transformers.feed_hotword(chunk)
                return True
        return False

    def _detect_ww(self, chunk: bytes) -> bool:
        """
        Check for a wake word in a chunk of unknown audio. Audio is passed to
        hotwords in all cases. If a wake word is detected
        audio is passed to `listenword_audio_callback` and `wake_callback`.

        If WW detected and sleeping, check for wakeup word in next audio chunks
        else check for speech input for STT.

        @param chunk:bytes of audio captured
        @return: True if a wakeword was detected
        """
        self.hotwords.state = HotwordState.LISTEN
        self.hotword_chunks.append(chunk)
        self.stt_chunks.append(chunk)
        self.hotwords.update(chunk)

        ww = self.hotwords.found()
        if ww:
            LOG.debug(f"Wake word detected={ww}")
            ww_data = self.hotwords.get_ww(ww)

            # Callback to handle recorded hotword audio
            if self.listenword_audio_callback is not None:
                hotword_audio_bytes = bytes()
                while self.hotword_chunks:
                    hotword_audio_bytes += self.hotword_chunks.popleft()

                self.listenword_audio_callback(hotword_audio_bytes, ww_data)

            self.hotword_chunks.clear()

            # Callback to handle wake up
            if self.wake_callback is not None:
                # emit record_begin
                self.wake_callback()

            if self.listen_mode == ListeningMode.SLEEPING:
                # Wake word detected, begin detecting "wake up" word
                self.state = ListeningState.CHECK_WAKE_UP
            else:
                if ww_data.get("sound"):
                    self.state = ListeningState.CONFIRMATION
                    # derive timeout from sound file length if possible
                    dur = ww_data.get("sound_duration", self.confirmation_seconds)
                    self.confirmation_seconds_left = dur
                else:
                    self.state = ListeningState.BEFORE_COMMAND
                # Wake word detected, begin recording voice command
                self.reset_speech_timer()
                self.stt_audio_bytes = bytes()
                self.stt.stream_start()
                if self.fallback_stt is not None:
                    self.fallback_stt.stream_start()

            LOG.debug(f"STATE: {self.state}")
            self.last_ww = time.time()
            self.transformers.feed_hotword(chunk)
            return True

        return False

    def _wait_cmd(self, chunk: bytes):
        """
        Handle audio chunks while in continuous listening mode, before VAD has
        detected any speech.
        @param chunk: bytes of audio captured
        """
        # Recording voice command, but user has not spoken yet
        self._chunk_info.is_speech = not self.vad.is_silence(chunk)
        hot = False
        if self._chunk_info.is_speech:
            self.speech_seconds_left -= self.mic.seconds_per_chunk
            if self.speech_seconds_left <= 0:
                # Voice command has started, so start looking for the end.
                if self.listen_mode == ListeningMode.CONTINUOUS:
                    prev_audio = len(self.stt_chunks) * self.mic.seconds_per_chunk
                    LOG.debug(f"waiting for speech: {prev_audio}")
                    self.stt.stream_start()
                    if self.fallback_stt is not None:
                        self.fallback_stt.stream_start()
                    self.state = ListeningState.IN_COMMAND
                else:
                    self.state = ListeningState.BEFORE_COMMAND
                LOG.debug(f"STATE: {self.state}")
        else:
            # Reset
            self.speech_seconds_left = self.speech_seconds
            # check hotwords
            hot = self._detect_hot(chunk)

        if not hot:
            self.transformers.feed_audio(chunk)
            if self.listen_mode == ListeningMode.CONTINUOUS:
                self.stt_audio_bytes += chunk
                self.stt_chunks.append(chunk)

    def _confirmation_sound(self, chunk: bytes):
        self._chunk_info.is_listen_sound = True
        if self.instant_listen:
            LOG.debug("instant_listen is on")
            self.confirmation_seconds_left = 0
            self.state = ListeningState.BEFORE_COMMAND
            LOG.debug(f"STATE: {self.state}")
            self._before_cmd(chunk)
            return

        # skip STT buffer if instant_listen is NOT set
        # Recording voice command, but user has not spoken yet
        self.transformers.feed_audio(chunk)
        self.confirmation_seconds_left -= self.mic.seconds_per_chunk
        if self.confirmation_seconds_left <= 0:
            self.state = ListeningState.BEFORE_COMMAND
            LOG.debug(f"STATE: {self.state}")

    def _before_cmd(self, chunk: bytes):
        """
        Handle audio chunks after WW detection or listen triggered, before VAD
        has detected any speech.
        @param chunk: bytes of audio captured
        """
        # Recording voice command, but user has not spoken yet
        self.transformers.feed_audio(chunk)

        self.stt_audio_bytes += chunk
        self.stt_chunks.append(chunk)
        while self.stt_chunks:
            stt_chunk = self.stt_chunks.popleft()
            self.stt.stream_data(stt_chunk)
            if self.fallback_stt is not None:
                self.fallback_stt.stream_data(stt_chunk)

            self.timeout_seconds_left -= self.mic.seconds_per_chunk
            self.timeout_seconds_with_silence_left -= self.mic.seconds_per_chunk
            if self.timeout_seconds_with_silence_left <= 0 or self.timeout_seconds_left <= 0:
                # Recording has timed out
                self.state = ListeningState.AFTER_COMMAND
                LOG.debug(f"STATE: {self.state}")
                break

            # Wait for enough speech before looking for the end of the
            # command (silence).
            try:
                self._chunk_info.is_speech = not self.vad.is_silence(stt_chunk)
            except Exception as e:
                LOG.exception(f"Error processing chunk of "
                              f"size={len(stt_chunk)} with "
                              f"SR={self.vad.sample_rate}: {e}")

            if self._chunk_info.is_speech:
                self.speech_seconds_left -= self.mic.seconds_per_chunk
                if self.speech_seconds_left <= 0:
                    # Voice command has started, so start looking for the
                    # end.
                    self.state = ListeningState.IN_COMMAND
                    self.silence_seconds_left = self.silence_seconds
                    LOG.debug(f"STATE: {self.state}")
                    break
            else:
                # Reset
                self.speech_seconds_left = self.speech_seconds

    def _in_cmd(self, chunk: bytes):
        """
        Handle audio chunks after VAD has identified speech and before the end
        of user speech is identified
        @param chunk: bytes of audio captured
        """
        self.transformers.feed_speech(chunk)

        # Recording voice command until user stops speaking
        self.stt_audio_bytes += chunk
        self.stt_chunks.append(chunk)
        while self.stt_chunks:
            stt_chunk = self.stt_chunks.popleft()

            self.stt.stream_data(stt_chunk)
            if self.fallback_stt is not None:
                self.fallback_stt.stream_data(stt_chunk)

            self.timeout_seconds_left -= self.mic.seconds_per_chunk
            if self.timeout_seconds_left <= 0:
                # Recording has timed out
                self.state = ListeningState.AFTER_COMMAND
                LOG.debug(f"STATE: {self.state}")
                break

            # Wait for enough silence before considering the command to be
            # ended.
            self._chunk_info.is_speech = not self.vad.is_silence(stt_chunk)
            if not self._chunk_info.is_speech:
                self.silence_seconds_left -= self.mic.seconds_per_chunk
                if self.silence_seconds_left <= 0:
                    # End of voice command detected
                    self.state = ListeningState.AFTER_COMMAND
                    LOG.debug(f"STATE: {self.state}")
                    break
            else:
                # Reset
                self.silence_seconds_left = self.silence_seconds

    def _validate_lang(self, lang: str) -> str:
        """
        ensure lang classification from speech is one of the valid langs
        if not then drop classification, as there are no speakers of that
        language around this device
        @param lang: BCP-47 language code to evaluate
        @return: validated language (or default)
        """
        default_lang = Configuration().get("lang", "en-us")
        valid_langs = [default_lang] + Configuration().get("secondary_langs", [])
        valid_langs = [l.lower().split("-")[0] for l in valid_langs]
        l2 = lang.lower().split("-")[0]
        if l2 in valid_langs:
            if l2 != default_lang.lower().split("-")[0]:
                LOG.info(f"replaced {default_lang} with {lang}")
                return lang
        else:
            LOG.warning(f"ignoring classification: {lang} is not in enabled "
                        f"languages: {valid_langs}")

        return default_lang

    def _get_tx(self, stt_context: dict) -> (str, dict):
        """
        Get a string transcription of audio that was previously streamed to STT.
        @param stt_context: dict context determined by transformers service
        @return: string transcription and dict context
        """
        # handle lang detection from speech
        if "stt_lang" in stt_context:
            lang = self._validate_lang(stt_context["stt_lang"])
            stt_context["stt_lang"] = lang
            # note: self.stt.stream is recreated every listen start
            # this is safe to do, and makes lang be passed to self.execute
            self.stt.stream.language = lang
            if self.fallback_stt:
                self.fallback_stt.stream.language = lang

        # get text and trigger callback
        try:
            utts = self.stt.transcribe() or []
        except:
            LOG.exception("STT failed")
            utts = []

        if not utts and self.fallback_stt is not None:
            LOG.info("Attempting fallback STT plugin")
            try:
                utts = self.fallback_stt.transcribe() or []
            except:
                LOG.exception("Fallback STT failed")

        if not utts:
            LOG.warning("STT transcription failed!")
            return [], stt_context

        filtered = [u for u in utts if u[1] >= self.min_stt_confidence]
        if not filtered:
            # ensure min 1 transcript
            filtered = [max(utts, key=lambda k: k[1])]
            LOG.warning("STT transcription below minimum confidence level!!!")

        if filtered != utts:
            LOG.info(f"Ignoring low confidence STT transcriptions: {[u for u in utts if u not in filtered]}")

        if len(filtered) > self.max_transcripts:
            LOG.debug(f"selecting top {self.max_transcripts} transcriptions")
            filtered = filtered[:self.max_transcripts]

        stt_context["transcriptions"] = filtered
        return filtered, stt_context

    def _vad_remove_silence(self):
        """removes silence from the STT buffer using the VAD plugin
        trimmed audio will never be < 1 second
        """
        # NOTE: This is using the FS-STT buffer directly, not the S-STT queue
        n_chunks = len(self.stt_audio_bytes) / self.mic.chunk_size
        seconds = n_chunks * self.mic.seconds_per_chunk
        LOG.debug(f"recorded {seconds} seconds of audio")
        if seconds > 1:
            extracted_speech = self.vad.extract_speech(self.stt_audio_bytes)
            if extracted_speech is None:
                LOG.debug("audio appears to be full silence! skipping VAD silence removal")
                return
            n_chunks = len(extracted_speech) / self.mic.chunk_size
            seconds2 = n_chunks * self.mic.seconds_per_chunk
            LOG.debug(f"removed {seconds - seconds2} seconds of silence, "
                      f"trimmed audio has {seconds2} seconds")
            if extracted_speech and seconds2 >= 1:
                self.stt.stream.buffer.clear()
                # replace the stt buffer with cropped audio
                self.stt.stream.update(extracted_speech)
            else:
                LOG.debug("trimmed audio is too short! skipping VAD silence removal")
        else:
            LOG.debug(f"skipping silence removal")

    def _after_cmd(self, chunk: bytes):
        """
        Handle audio chunk after VAD has determined a command is ended.
        Calls `stt_audio_callback` and `text_callback` with finalized utterance
        recording and transcript. The loop is reset to the appropriate state for
        the next command.
        @param chunk: bytes of audio captured
        """
        # Command has ended, call transformers pipeline before STT
        chunk, stt_context = self.transformers.transform(chunk)
        if isinstance(self.stt, FakeStreamingSTT) and self.remove_silence:
            self._vad_remove_silence()

        utts, stt_context = self._get_tx(stt_context)

        if utts:
            LOG.debug(f"transformers metadata: {stt_context}")
            LOG.info(f"transcribed: {utts}")
        else:
            LOG.info("nothing transcribed")
        # Voice command has finished recording
        if self.stt_audio_callback is not None:
            self.stt_audio_callback(self.stt_audio_bytes, stt_context)

        self.stt_audio_bytes = bytes()

        if self.record_end_callback is not None:
            # emit record_end
            self.record_end_callback()

        # Callback to handle STT text
        if self.text_callback is not None:
            self.text_callback(utts, stt_context)

        # Back to detecting wake word
        if self.listen_mode == ListeningMode.CONTINUOUS or \
                self.listen_mode == ListeningMode.HYBRID:
            self.state = ListeningState.WAITING_CMD
        else:
            self.state = ListeningState.DETECT_WAKEWORD
        LOG.debug(f"STATE: {self.state}")

        # Clear any buffered STT chunks
        self.stt_chunks.clear()

        # Reset wakeword detector state, if available
        self.hotwords.reset()

        # Reset the VAD internal state to avoid the model getting
        # into a degenerative state where it always reports silence.
        if hasattr(self.vad, "reset"):
            LOG.debug("reset VAD")
            self.vad.reset()

        self.timeout_seconds_left = self.timeout_seconds
        self.timeout_seconds_with_silence_left = self.timeout_seconds_with_silence
        
    def stop(self):
        """
        Signal the VoiceLoop to stop processing audio.
        """
        self._is_running = False



from typing import Any, Dict, Optional, List, Tuple

from ovos_config.config import Configuration
from ovos_plugin_manager.stt import OVOSSTTFactory
from ovos_plugin_manager.templates.stt import StreamingSTT, StreamThread
from ovos_plugin_manager.utils import ReadWriteStream
from ovos_utils.log import LOG
from speech_recognition import AudioData


class FakeStreamThread(StreamThread):

    def __init__(self, queue, language, engine, sample_rate, sample_width):
        super().__init__(queue, language)
        self.buffer = ReadWriteStream()
        self.engine = engine
        self.sample_rate = sample_rate
        self.sample_width = sample_width

    def finalize(self):
        """ return final transcription """

        if not self.buffer:
            return ""

        try:
            # plugins expect AudioData objects
            audio = AudioData(self.buffer.read(),
                              sample_rate=self.sample_rate,
                              sample_width=self.sample_width)
            transcript = self.engine.execute(audio, self.language)

            self.buffer.clear()
            return transcript
        except Exception:
            LOG.exception(f"Error in STT plugin: {self.engine.__class__.__name__}")
        return None

    def handle_audio_stream(self, audio, language):
        for chunk in audio:
            self.update(chunk)

    def update(self, chunk: bytes):
        self.buffer.write(chunk)


class FakeStreamingSTT(StreamingSTT):
    def __init__(self, engine, config=None):
        super().__init__(config)
        self.engine = engine

    def create_streaming_thread(self):
        listener = Configuration().get("listener", {})
        sample_rate = listener.get("sample_rate", 16000)
        sample_width = listener.get("sample_width", 2)
        return FakeStreamThread(self.queue, self.lang, self.engine, sample_rate,
                                sample_width)

    def transcribe(self, audio: Optional = None,
                   lang: Optional[str] = None) -> List[Tuple[str, float]]:
        """transcribe audio data to a list of
        possible transcriptions and respective confidences"""
        # plugins expect AudioData objects
        audiod = AudioData(audio or self.stream.buffer.read(),
                           sample_rate=self.stream.sample_rate,
                           sample_width=self.stream.sample_width)
        transcripts = self.engine.transcribe(audiod, lang)
        if audio is None:
            self.stream.buffer.clear()
        return transcripts


def load_stt_module(config: Dict[str, Any] = None) -> StreamingSTT:
    """
    Load an STT module based on configuration
    @param config: STT or global configuration or None (uses Configuration)
    @return: Initialized StreamingSTT plugin
    """
    # Create a copy because we're setting default values here
    stt_config = config or Configuration().get("stt", {})
    stt_config = dict(stt_config)
    default_lang = Configuration().get('lang')
    stt_config.setdefault("lang", default_lang)
    if stt_config['lang'] != default_lang:
        LOG.warning(f"STT lang ({stt_config['lang']} differs from global "
                    f"({Configuration.get('lang')}")
    plug = OVOSSTTFactory.create(stt_config)
    if not isinstance(plug, StreamingSTT):
        LOG.debug(f"Using FakeStreamingSTT wrapper with config={config}")
        return FakeStreamingSTT(plug, config)
    return plug


def load_fallback_stt(cfg: Dict[str, Any] = None) -> Optional[StreamingSTT]:
    """
    Load a fallback STT module based on configuration
    @param cfg: STT or global configuration or None (uses Configuration)
    @return: Initialized StreamingSTT plugin if configured, else None
    """
    cfg = cfg or Configuration().get("stt", {})
    default_lang = Configuration().get('lang')
    fbm = cfg.get("fallback_module")
    if not fbm:
        return None
    try:
        config = cfg.get(fbm, {})
        config.setdefault("lang", default_lang)
        if config['lang'] != default_lang:
            LOG.warning(f"Fallback STT lang ({config['lang']} differs from "
                        f"global ({Configuration.get('lang')}")
        plug = OVOSSTTFactory.create({"module": fbm, fbm: config})
        if not isinstance(plug, StreamingSTT):
            LOG.debug(f"Using FakeStreamingSTT wrapper with config={config}")
            return FakeStreamingSTT(plug, config)
        return plug
    except:
        LOG.exception("Failed to load fallback STT")
        return None



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Framework
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2022 Neongecko.com Inc.
# Contributors: Daniel McKnight, Guy Daniels, Elon Gasper, Richard Leeds,
# Regina Bloomstine, Casimiro Ferreira, Andrii Pernatii, Kirill Hrymailo
# BSD-3 License
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from ovos_plugin_manager.audio_transformers import find_audio_transformer_plugins
from ovos_utils.json_helper import merge_dict
from ovos_utils.log import LOG


class AudioTransformersService:

    def __init__(self, bus, config=None):
        self.config_core = config or {}
        self.loaded_plugins = {}
        self.has_loaded = False
        self.bus = bus
        # to activate a plugin, just add an entry to mycroft.conf for it
        self.config = \
            self.config_core.get("listener").get("audio_transformers") or dict()
        self.load_plugins()

    def load_plugins(self):
        for plug_name, plug in find_audio_transformer_plugins().items():
            if plug_name in self.config:
                # if disabled skip it
                if not self.config[plug_name].get("active", True):
                    continue
                try:
                    self.loaded_plugins[plug_name] = plug()
                    self.loaded_plugins[plug_name].bind(self.bus)
                    LOG.info(f"loaded audio transformer plugin: {plug_name}")
                except Exception as e:
                    LOG.exception(f"Failed to load audio transformer plugin: "
                                  f"{plug_name}")
        self.has_loaded = True

    @property
    def plugins(self) -> list:
        """
        Return loaded transformers in priority order, such that modules with a
        higher `priority` rank are called first and changes from lower ranked
        transformers are applied last.

        A plugin of `priority` 1 will override any existing context keys and
        will be the last to modify `audio_data`
        """
        return sorted(self.loaded_plugins.values(),
                      key=lambda k: k.priority, reverse=True)

    def shutdown(self):
        """
        Shutdown all loaded plugins
        """
        for module in self.plugins:
            try:
                module.shutdown()
            except Exception as e:
                LOG.warning(e)

    def feed_audio(self, chunk: bytes):
        """
        Feed a chunk of untagged (not speech) audio to all loaded plugins
        @param chunk: bytes of audio data
        """
        for module in self.plugins:
            module.feed_audio_chunk(chunk)

    def feed_hotword(self, chunk: bytes):
        """
        Feed a chunk of hotword audio to all loaded plugins
        @param chunk: bytes of audio data
        """
        for module in self.plugins:
            module.feed_hotword_chunk(chunk)

    def feed_speech(self, chunk: bytes):
        """
        Feed a chunk of speech audio to all loaded plugins
        @param chunk: bytes of audio data
        """
        try:
            for module in self.plugins:
                module.feed_speech_chunk(chunk)
        except Exception as e:
            LOG.exception(e)

    def transform(self, chunk: bytes) -> (bytes, dict):
        """
        Get transformed audio and context for the preceding audio
        @param chunk: bytes of audio data
        @return: transformed audio data, dict context
        """
        context = {'client_name': 'ovos_dinkum_listener',
                   'source': 'audio',  # default native audio source
                   'destination': ["skills"]}
        for module in self.plugins:
            try:
                LOG.debug(f"checking audio transformer: {module}")
                chunk = module.feed_speech_utterance(chunk)
                chunk, data = module.transform(chunk)
                LOG.debug(f"{module.name}: {data}")
                context = merge_dict(context, data)
            except Exception as e:
                LOG.exception(e)
        return chunk, context



from enum import Enum
from os.path import dirname
from threading import Event
from typing import Optional

from ovos_config import Configuration
from ovos_plugin_manager.wakewords import OVOSWakeWordFactory, HotWordEngine
from ovos_utils.fakebus import FakeBus
from ovos_utils.log import LOG
try:
    from ovos_utils.sound import get_sound_duration
except ImportError:

    def get_sound_duration(*args, **kwargs):
        raise ImportError("please install ovos-utils>=0.1.0a25")


class HotWordException(RuntimeWarning):
    """Exception related to HotWords"""


class CyclicAudioBuffer:
    def __init__(self, duration=0.98, initial_data=None,
                 sample_rate=16000, sample_width=2):
        self.size = self.duration_to_bytes(duration, sample_rate, sample_width)
        initial_data = initial_data or self.get_silence(self.size)
        # Get at most size bytes from the end of the initial data
        self._buffer = initial_data[-self.size:]

    def clear(self):
        """
        Set the buffer to empty data
        """
        self._buffer = self.get_silence(self.size)

    @staticmethod
    def duration_to_bytes(duration: float, sample_rate: int = 16000,
                          sample_width: int = 2) -> int:
        """
        Convert duration in seconds to a number of bytes
        @param duration: duration in seconds
        @param sample_rate: sample rate of expected audio
        @param sample_width: sample width of expected audio
        @return: number of bytes
        """
        return int(duration * sample_rate) * sample_width

    @staticmethod
    def get_silence(num_bytes: int) -> bytes:
        """
        Return null bytes
        @param num_bytes: number of bytes to return
        @return: requested number of null bytes
        """
        return b'\0' * num_bytes

    def append(self, data: bytes):
        """
        Add new data to the buffer, and slide out data if the buffer is full
        @param data: binary data to append to the buffer.
            If buffer size is exceeded, the oldest data will be dropped.
        """
        buff = self._buffer + data
        if len(buff) > self.size:
            buff = buff[-self.size:]
        self._buffer = buff

    def get(self) -> bytes:
        """
        Get the binary audio data from the buffer
        """
        return self._buffer


class HotwordState(str, Enum):
    """ current listener state """
    LISTEN = "wakeword"
    HOTWORD = "hotword"
    RECORDING = "recording"
    WAKEUP = "wakeup"


def _safe_get_plugins(func):
    def wrapped(*args, **kwargs):
        if not HotwordContainer._loaded.wait(30):
            raise TimeoutError("Timed out waiting for Hotwords load")
        try:
            return func(*args, **kwargs)
        except KeyError:
            raise HotWordException("Expected engine not loaded")

    return wrapped


class HotwordContainer:
    _plugins = {}
    _loaded = Event()

    def __init__(self, bus=FakeBus(), expected_duration=3, sample_rate=16000,
                 sample_width=2):
        self.bus = bus
        self.state = HotwordState.HOTWORD
        # used for old style non-streaming wakeword (deprecated)
        self.audio_buffer = CyclicAudioBuffer(expected_duration,
                                              sample_rate=sample_rate,
                                              sample_width=sample_width)
        self.reload_on_failure = False
        self.applied_hotwords_config = None

    def load_hotword_engines(self):
        """
        Load hotword objects from configuration
        """
        self._loaded.clear()
        LOG.info("creating hotword engines")
        config_core = Configuration()
        default_lang = config_core.get("lang", "en-us")
        hot_words = config_core.get("hotwords", {})
        self.applied_hotwords_config = hot_words
        global_listen = config_core.get("confirm_listening")
        global_sounds = config_core.get("sounds", {})

        main_ww = config_core.get("listener",
                                  {}).get("wake_word",
                                          "hey_mycroft").replace(" ", "_")
        wakeupw = config_core.get("listener",
                                  {}).get("stand_up_word",
                                          "wake_up").replace(" ", "_")

        for word, data in dict(hot_words).items():
            try:
                # normalization step to avoid naming collisions
                # TODO - move this to ovos_config package,
                #  on changes to the hotwords section this should be enforced directly
                # this approach does not fully solve the issue, config merging may be messed up
                word = word.replace(" ", "_")

                sound = data.get("sound")
                utterance = data.get("utterance")
                listen = data.get("listen", False) or word == main_ww
                wakeup = data.get("wakeup", False)
                stopword = data.get("stopword", False)
                lang = data.get("stt_lang", default_lang)
                enabled = data.get("active")
                event = data.get("bus_event")

                # automatically enable default wake words
                # only if the active status is undefined
                if enabled is None:
                    if word == main_ww or word == wakeupw:
                        enabled = True
                    else:
                        enabled = False

                # global listening sound
                if not sound and listen and global_listen:
                    sound = global_sounds.get("start_listening")

                if not enabled:
                    continue

                engine = OVOSWakeWordFactory.create_hotword(word, lang=lang)
                if engine is not None:
                    LOG.info(f"Loading hotword: {word} with engine: {engine}")
                    if hasattr(engine, "bind"):
                        engine.bind(self.bus)
                        # not all plugins implement this
                    if data.get('engine'):
                        LOG.info(f"Engine previously defined. "
                                 f"Deleting old instance.")
                        try:
                            data['engine'].stop()
                            del data['engine']
                        except Exception as e:
                            LOG.error(e)
                    self._plugins[word] = {"engine": engine,
                                           "sound": sound,
                                           "bus_event": event,
                                           "utterance": utterance,
                                           "stt_lang": lang,
                                           "listen": listen,
                                           "wakeup": wakeup,
                                           "stopword": stopword}
                    if sound:
                        try:
                            if sound.startswith("snd/"):
                                dur = get_sound_duration(sound,
                                                         base_dir=f"{dirname(dirname(__file__))}/res")
                            else:
                                dur = get_sound_duration(sound)
                            LOG.debug(f"{sound} duration: {dur} seconds")
                            self._plugins[word]["sound_duration"] = dur
                        except:
                            pass

            except Exception as e:
                LOG.error("Failed to load hotword: " + word)

        self._loaded.set()

        if not self.listen_words:
            LOG.error("No listen words loaded")
        else:
            self.reload_on_failure = True
        if not self.wakeup_words:
            LOG.warning("No wakeup words loaded")
        if not self.stop_words:
            LOG.warning("No stop words loaded")

    @property
    def ww_names(self):
        """ wakeup words exit sleep mode if detected after a listen word"""
        return list(self._plugins.keys())

    @property
    @_safe_get_plugins
    def plugins(self):
        return [v["engine"] for k, v in self._plugins.items()]

    @property
    @_safe_get_plugins
    def wakeup_words(self):
        """ wakeup words exit sleep mode if detected after a listen word"""
        return {k: v["engine"] for k, v in self._plugins.items()
                if v.get("wakeup")}

    @property
    @_safe_get_plugins
    def listen_words(self):
        """ listen words trigger the VAD/STT stages"""
        return {k: v["engine"] for k, v in self._plugins.items()
                if v.get("listen")}

    @property
    @_safe_get_plugins
    def stop_words(self):
        """ stop only work during recording mode, they exit recording mode"""
        return {k: v["engine"] for k, v in self._plugins.items()
                if v.get("stopword")}

    @property
    @_safe_get_plugins
    def hot_words(self):
        """ hotwords only emit bus events / play sounds, they do not affect listening loop"""
        return {k: v["engine"] for k, v in self._plugins.items()
                if not v.get("stopword") and
                not v.get("wakeup") and
                not v.get("listen")}

    def found(self) -> Optional[str]:
        """
        Check if a hotword is found in a relevant engine, based on self.state
        @return: string detected hotword, else None
        """
        # Check for which detectors we want; if none are active, log something
        # because it means there's no ww that "exits" the current state
        if self.state == HotwordState.LISTEN:
            engines = self.listen_words
            if not engines:
                raise HotWordException(
                    f"Waiting for listen_words but none are available!")
        elif self.state == HotwordState.WAKEUP:
            engines = self.wakeup_words
        elif self.state == HotwordState.RECORDING:
            engines = self.stop_words
        else:
            engines = self.hot_words

        # streaming engines will ignore the byte_data
        audio_data = self.audio_buffer.get()
        for ww_name, engine in engines.items():
            try:
                assert isinstance(engine, HotWordEngine)
                # non-streaming ww engines expect a 3-second cyclic buffer here
                # streaming engines will ignore audio_data
                # (got it via self.update)
                if engine.found_wake_word(audio_data):
                    LOG.debug(f"Detected wake_word: {ww_name}")
                    return ww_name
            except AssertionError:
                LOG.error(f"Expected HotWordEngine, but got: {engine} for "
                          f"{ww_name}")
                # TODO: Add engine reload here?
            except Exception as e:
                LOG.error(e)
        return None

    def get_ww(self, ww: str) -> dict:
        """
        Get information about the requested wake word
        @param ww: string wake word to get information for
        @return: dict wake word information
        """
        if ww not in self._plugins:
            raise ValueError(f"Requested ww not defined: {ww}")
        meta = dict(self._plugins.get(ww))
        plug = meta["engine"]
        assert isinstance(plug, HotWordEngine)
        meta["key_phrase"] = ww
        meta["module"] = plug.config["module"]
        meta["engine"] = plug.__class__.__name__
        return meta

    def update(self, chunk: bytes):
        """
        Update appropriate engines based on self.state
        @param chunk: bytes of audio to feed to hotword engines
        """
        self.audio_buffer.append(chunk)
        if self.state == HotwordState.LISTEN:
            # LOG.debug(f"Update listen_words")
            engines = self.listen_words.values()
        elif self.state == HotwordState.WAKEUP:
            # LOG.debug(f"Update wakeup_words")
            engines = self.wakeup_words.values()
        elif self.state == HotwordState.RECORDING:
            # LOG.debug(f"Update stop_words")
            engines = self.stop_words.values()
        else:
            # LOG.debug(f"Update hot_words")
            engines = self.hot_words.values()

        for engine in engines:
            try:
                # old style engines will ignore the update
                engine.update(chunk)
            except Exception as e:
                LOG.error(e)

    def reset(self):
        """
        Clear the audio_buffer and reset all hotword engines
        """
        self.audio_buffer.clear()
        for engine in self.plugins:
            try:
                # TODO: Remove check when default method is added to base class
                if hasattr(engine, 'reset'):
                    engine.reset()
            except Exception as e:
                LOG.error(e)

    def shutdown(self):
        """
        Shutdown all engines, remove references to plugins
        """
        for engine in self.plugins:
            try:
                engine.shutdown()
            except Exception as e:
                LOG.error(e)
        for ww in self.ww_names:
            self._plugins.pop(ww)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from ovos_utils.log import init_service_logger

from ovos_dinkum_listener.service import OVOSDinkumVoiceService, on_ready, \
    on_error, on_stopping


def main(ready_hook=on_ready, error_hook=on_error, stopping_hook=on_stopping,
         watchdog=lambda: None):
    """Service entry point"""
    init_service_logger("voice")
    service = OVOSDinkumVoiceService(on_ready=ready_hook, on_error=error_hook,
                                     on_stopping=stopping_hook,
                                     watchdog=watchdog)
    service.run()
    # wait until services exits by itself, or crashes


if __name__ == "__main__":
    main()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import base64
import json
import subprocess
import time
import wave
from enum import Enum
from hashlib import md5
from os.path import dirname
from pathlib import Path
from tempfile import NamedTemporaryFile
from threading import Thread, RLock, Event
from typing import List, Tuple

import speech_recognition as sr
from distutils.spawn import find_executable
from ovos_bus_client import MessageBusClient
from ovos_bus_client.message import Message
from ovos_bus_client.session import SessionManager
from ovos_config import Configuration
from ovos_config.locations import get_xdg_data_save_path
from ovos_plugin_manager.microphone import OVOSMicrophoneFactory
from ovos_plugin_manager.stt import get_stt_lang_configs, get_stt_supported_langs, get_stt_module_configs
from ovos_plugin_manager.utils.tts_cache import hash_sentence
from ovos_plugin_manager.vad import OVOSVADFactory
from ovos_plugin_manager.vad import get_vad_configs
from ovos_plugin_manager.wakewords import get_ww_lang_configs, get_ww_supported_langs, get_ww_module_configs
from ovos_utils.log import LOG, log_deprecation
from ovos_utils.process_utils import ProcessStatus, StatusCallbackMap, ProcessState

from ovos_dinkum_listener.plugins import load_stt_module, load_fallback_stt
from ovos_dinkum_listener.transformers import AudioTransformersService
from ovos_dinkum_listener.voice_loop import DinkumVoiceLoop, ListeningMode, ListeningState
from ovos_dinkum_listener.voice_loop.hotwords import HotwordContainer
try:
    from ovos_backend_client.api import DatasetApi
except ImportError:
    LOG.info("`ovos-backend-client` is not installed. Upload is disabled")
    DatasetApi = None

try:
    from ovos_utils.sound import get_sound_duration
except ImportError:

    def get_sound_duration(*args, **kwargs):
        raise ImportError("please install ovos-utils>=0.1.0a25")

# Seconds between systemd watchdog updates
WATCHDOG_DELAY = 0.5


def bytes2audiodata(data):
    recognizer = sr.Recognizer()
    with NamedTemporaryFile() as fp:
        fp.write(data)

        if find_executable("ffmpeg"):
            p = fp.name + "converted.wav"
            # ensure file format
            cmd = ["ffmpeg", "-i", fp.name, "-acodec", "pcm_s16le", "-ar",
                   "16000", "-ac", "1", "-f", "wav", p, "-y"]
            subprocess.call(cmd)
        else:
            LOG.warning("ffmpeg not found, please ensure audio is in a valid format")
            p = fp.name

        with sr.AudioFile(p) as source:
            audio = recognizer.record(source)
    return audio


class ServiceState(str, Enum):
    NOT_STARTED = "not_started"
    STARTED = "started"
    RUNNING = "running"
    STOPPING = "stopping"


def on_ready():
    LOG.info('DinkumVoiceService is ready.')


def on_alive():
    LOG.info('DinkumVoiceService is alive.')


def on_started():
    LOG.info('DinkumVoiceService started.')


def on_error(e='Unknown'):
    LOG.error(f'DinkumVoiceService failed to launch ({e}).')


def on_stopping():
    LOG.info('DinkumVoiceService is shutting down...')


class OVOSDinkumVoiceService(Thread):
    """
    Service for handling user voice input.

    Performs the following tasks:
    * Recording audio from microphone
    * Hotword detection
    * Voice activity detection (silence at end of voice command)
    * Speech to text

    Input messages:
    * mycroft.mic.mute
      * Produces empty audio stream
    * mycroft.mic.unmute
      * Uses real audio stream
    * mycroft.mic.listen
      * Wakes up mycroft and starts recording voice command

    Output messages:
    * recognizer_loop:awoken
      * Reports that mycroft is now awake
    * recognizer_loop:wake
      * Reports wake word used to wake up mycroft
    * recognizer_loop:record_begin
      * Reports that voice command recording has begun
    * recognizer_loop:record_end
      * Reports that voice command recording has ended
    * recognizer_loop:utterance
      * Result from speech to text of voice command
    * recognizer_loop:speech.recognition.unknown
      * Sent when empty result from speech to text is returned

    Service messages:
    * voice.service.connected
    * voice.service.connected.response
    * voice.initialize.started
    * voice.initialize.ended

    """

    def __init__(self, on_ready=on_ready, on_error=on_error,
                 on_stopping=on_stopping, on_alive=on_alive,
                 on_started=on_started, watchdog=lambda: None, mic=None,
                 bus=None, validate_source=True, *args, **kwargs):
        """
        watchdog: (callable) function to call periodically indicating
          operational status.
        """
        super().__init__(*args, **kwargs)

        LOG.info("Starting Voice Service")
        callbacks = StatusCallbackMap(on_ready=on_ready,
                                      on_error=on_error,
                                      on_stopping=on_stopping,
                                      on_alive=on_alive,
                                      on_started=on_started)
        self.bus = bus
        self.service_id = "voice"
        self.validate_source = validate_source
        self.status = ProcessStatus(self.service_id, self.bus,
                                    callback_map=callbacks)
        self._watchdog = watchdog
        self._shutdown_event = Event()
        self._stopping = False
        self.status.set_alive()
        self.config = Configuration()
        self._applied_config_hash = self._config_hash()
        self._default_vol = 70  # for barge-in

        self._before_start()  # connect to bus

        # Initialize with default (bundled) plugin
        microphone_config = self.config.get("listener",
                                            {}).get("microphone", {})
        microphone_config.setdefault('module', 'ovos-microphone-plugin-alsa')

        self.mic = mic or OVOSMicrophoneFactory.create(microphone_config)

        self.hotwords = HotwordContainer(self.bus)
        self.vad = OVOSVADFactory.create()
        self.stt = load_stt_module()
        self.fallback_stt = load_fallback_stt()
        self.transformers = AudioTransformersService(self.bus, self.config)

        self._load_lock = RLock()
        self._reload_event = Event()
        self._reload_event.set()
        listener = self.config["listener"]
        self.voice_loop = self._init_voice_loop(listener)

    def _validate_message_context(self, message, native_sources=None):
        """ used to determine if a message should be processed or ignored
        only native sources should trigger on mycroft.mic.listen
        """
        if not message or not self.validate_source:
            return True
        destination = message.context.get("destination")
        if destination:
            native_sources = native_sources or \
                             Configuration()["Audio"].get("native_sources",
                                                          ["debug_cli", "audio"]) or []
            if any(s in destination for s in native_sources):
                # request from device
                return True
            # external request, do not handle
            return False
        # broadcast for everyone
        return True

    def _config_hash(self):
        lang = self.config.get("lang")
        stt_module = self.config.get("stt", {}).get("module")
        fallback_module = self.config.get("stt", {}).get("fallback_module")
        stt_config = {
            "lang": lang,
            "module": stt_module,
            "config": self.config.get("stt", {}).get(stt_module)
        }
        stt_fallback = {
            "lang": lang,
            "module": fallback_module,
            "config": self.config.get("stt", {}).get(fallback_module)
        }
        loop_config = {
            "listener": self.config.get('listener'),
            "vad": self.config.get("VAD"),
            "mic": self.config.get("microphone")
        }
        hotword_config = {
            "confirm": self.config.get('confirm_listening'),
            "sounds": self.config.get("sounds"),
            "listener": self.config.get('listener'),  # Defaults, save_hotwords
            "hotwords": self.config.get('hotwords')
        }
        config_hashes = {
            "loop": hash(json.dumps(loop_config)),
            "hotwords": hash(json.dumps(hotword_config)),
            "stt": hash(json.dumps(stt_config)),
            "fallback": hash(json.dumps(stt_fallback)),
        }
        return config_hashes

    def _init_voice_loop(self, listener_config: dict):
        """
        Initialize a DinkumVoiceLoop object with the specified config
        @param listener_config:
        @return: Initialized VoiceLoop object
        """
        with self._load_lock:
            loop = DinkumVoiceLoop(
                mic=self.mic,
                hotwords=self.hotwords,
                stt=self.stt,
                fallback_stt=self.fallback_stt,
                vad=self.vad,
                transformers=self.transformers,
                instant_listen=listener_config.get("instant_listen", True),
                speech_seconds=listener_config.get("speech_begin", 0.3),
                silence_seconds=listener_config.get("silence_end", 0.7),
                timeout_seconds=listener_config.get("recording_timeout", 10),
                timeout_seconds_with_silence=listener_config.get("recording_timeout_with_silence", 5),
                recording_mode_max_silence_seconds=listener_config.get("recording_mode_max_silence_seconds", 30),
                num_stt_rewind_chunks=listener_config.get("utterance_chunks_to_rewind", 2),
                num_hotword_keep_chunks=listener_config.get("wakeword_chunks_to_save", 15),
                remove_silence=listener_config.get("remove_silence", False),
                wake_callback=self._record_begin,
                text_callback=self._stt_text,
                listenword_audio_callback=self._hotword_audio,
                hotword_audio_callback=self._hotword_audio,
                stopword_audio_callback=self._hotword_audio,
                wakeupword_audio_callback=self._hotword_audio,
                stt_audio_callback=self._stt_audio,
                recording_audio_callback=self._recording_audio,
                wakeup_callback=self._wakeup,
                record_end_callback=self._record_end_signal,
                min_stt_confidence=listener_config.get("min_stt_confidence", 0.6),
                max_transcripts=listener_config.get("max_transcripts", 1)
            )
        return loop

    @property
    def default_save_path(self):
        """ where recorded hotwords/utterances are saved """
        listener = Configuration().get("listener", {})
        return listener.get('save_path', f"{get_xdg_data_save_path()}/listener")

    @property
    def state(self):
        log_deprecation("This property is deprecated, reference `status.state`",
                        "0.1.0")
        if self.status.state in (ProcessState.NOT_STARTED, ProcessState.ALIVE):
            return ServiceState.NOT_STARTED
        if self.status.state == ProcessState.STARTED:
            return ServiceState.STARTED
        if self.status.state == ProcessState.READY:
            return ServiceState.RUNNING
        if self.status.state in (ProcessState.ERROR, ProcessState.STOPPING):
            return ServiceState.STOPPING
        return self.status.state

    def run(self):
        """
        Service entry point
        """
        try:
            self.status.set_alive()
            self._start()
            self.status.set_started()
            self._after_start()
            LOG.debug("Service started")
        except Exception as e:
            LOG.exception("Service failed to start")
            self.status.set_error(str(e))
            return

        try:
            self.status.set_ready()
            LOG.info("Service ready")
            while not self._stopping:
                if not self._reload_event.wait(30):
                    raise TimeoutError("Timed out waiting for reload")
                self.voice_loop.run()
        except KeyboardInterrupt:
            LOG.info("Exit via CTRL+C")
        except Exception as e:
            LOG.exception("voice_loop failed")
            self.status.set_error(str(e))
        finally:
            LOG.info("Service stopping")
            self.stop()
            LOG.debug("shutdown done")
            LOG.debug("stopped")
            if self.status.state != ProcessState.ERROR:
                self.status.state = ProcessState.NOT_STARTED

    def _before_start(self):
        """
        Initialization logic called on module init
        """
        self.config = self.config or Configuration()
        LOG.info("Starting service...")
        self._connect_to_bus()

    def _start(self):
        """
        Start microphone and listener loop
        @return:
        """
        self.mic.start()
        self.hotwords.load_hotword_engines()
        self.voice_loop.start()
        self.register_event_handlers()

    def register_event_handlers(self):
        # Register events
        self.bus.on("mycroft.mic.mute", self._handle_mute)
        self.bus.on("mycroft.mic.unmute", self._handle_unmute)
        self.bus.on("mycroft.mic.mute.toggle", self._handle_mute_toggle)

        self.bus.on("mycroft.mic.listen", self._handle_listen)
        self.bus.on('mycroft.mic.get_status', self._handle_mic_get_status)
        self.bus.on('recognizer_loop:audio_output_start', self._handle_audio_start)
        self.bus.on('recognizer_loop:audio_output_end', self._handle_audio_end)
        self.bus.on('mycroft.stop', self._handle_stop)

        self.bus.on('recognizer_loop:sleep', self._handle_sleep)
        self.bus.on('recognizer_loop:wake_up', self._handle_wake_up)
        self.bus.on('recognizer_loop:b64_audio', self._handle_b64_audio)
        self.bus.on('recognizer_loop:record_stop', self._handle_stop_recording)
        self.bus.on('recognizer_loop:state.set', self._handle_change_state)
        self.bus.on('recognizer_loop:state.get', self._handle_get_state)
        self.bus.on("intent.service.skills.activated", self._handle_extend_listening)

        self.bus.on("ovos.languages.stt", self._handle_get_languages_stt)
        self.bus.on("opm.stt.query", self._handle_opm_stt_query)
        self.bus.on("opm.ww.query", self._handle_opm_ww_query)
        self.bus.on("opm.vad.query", self._handle_opm_vad_query)

        self.bus.on("mycroft.audio.play_sound.response", self._handle_sound_played)

        # tracking volume for fake barge-in
        self.bus.on("volume.set.percent", self._handle_volume_change)
        self.bus.on("mycroft.volume.increase", self._handle_volume_change)
        self.bus.on("mycroft.volume.decrease", self._handle_volume_change)
        self._query_volume()  # sync initial volume state

        LOG.debug("Messagebus events registered")

    def _after_start(self):
        """Initialization logic called after start()"""
        Thread(target=self._pet_the_dog, daemon=True).start()
        self.status.set_started()

    def stop(self):
        """
        Stop the voice_loop and trigger service shutdown
        """
        self.status.set_stopping()
        self._stopping = True
        if self.voice_loop.running:
            self.voice_loop.stop()
        self._shutdown()

    def _shutdown(self):
        """
        Internal method to shut down any running services. Called after
        `self.voice_loop` is stopped
        """
        if not self._load_lock.acquire(timeout=30):
            LOG.warning("Lock not acquired after 30 seconds; "
                        "shutting down anyways")
        try:
            if hasattr(self.stt, "shutdown"):
                self.stt.shutdown()

            if hasattr(self.fallback_stt, "shutdown"):
                self.fallback_stt.shutdown()

            self.hotwords.shutdown()

            if hasattr(self.vad, "stop"):
                self.vad.stop()

            self.mic.stop()

            self.bus.close()
        except Exception as e:
            LOG.exception(f"Shutdown failed with: {e}")
        self._shutdown_event.set()
        self._load_lock.release()

    def _connect_to_bus(self):
        """Connects to the websocket message bus"""
        self.bus = self.bus or MessageBusClient()
        if not self.bus.started_running:
            LOG.debug("Starting bus connection")
            self.bus.run_in_thread()
            self.bus.connected_event.wait()
        if not self.status.bus:
            self.status.bind(self.bus)
        self.config.set_config_update_handlers(self.bus)
        self.config.set_config_watcher(self.reload_configuration)
        LOG.info("Connected to Mycroft Core message bus")

    def _report_service_state(self, message):
        """Response to service state requests"""
        self.bus.emit(message.response(data={"state": self.state.value})),

    def _pet_the_dog(self):
        """Notify systemd that the service is still running"""
        if self._watchdog is not None:
            try:
                while True:
                    # Prevent systemd from restarting service
                    self._watchdog()
                    time.sleep(WATCHDOG_DELAY)
            except Exception:
                LOG.exception("Unexpected error in watchdog thread")

    # Fake Barge In
    def _query_volume(self):
        """get the default volume"""
        response = self.bus.wait_for_response(Message("mycroft.volume.get"))
        if response:
            self._default_vol = int(response.data["percent"] * 100)

    @property
    def fake_barge_in(self) -> bool:
        """lower volume during recording"""
        return Configuration().get("listener", {}).get("fake_barge_in", False)

    @property
    def fake_barge_in_volume(self) -> int:
        """volume to set when recording"""
        return Configuration().get("listener", {}).get("barge_in_volume", 30)

    def _handle_volume_change(self, message: Message):
        """keep track of volume changes so we restore to the correct level"""
        if not self.fake_barge_in or message.context.get("skill_id", "") == "dinkum-listener":
            # ignore our own messages
            return
        if message.msg_type == "mycroft.volume.increase":
            vol = int(message.data.get("percent", .1) * 100)
            self._default_vol += vol
        elif message.msg_type == "mycroft.volume.decrease":
            vol = int(message.data.get("percent", -.1) * 100)
            self._default_vol -= abs(vol)
        else:
            vol = int(message.data["percent"] * 100)
            self._default_vol = vol
        LOG.info(f"tracking user volume for after barge-in: {self._default_vol}")

    # callbacks
    def _wakeup(self):
        """ callback when voice loop exits SLEEP mode"""
        self.bus.emit(Message("mycroft.awoken"))

    def _record_begin(self):
        LOG.debug("Record begin")
        if self.fake_barge_in:
            LOG.info(f"fake barge-in lowering volume to: {self.fake_barge_in_volume}")
            self.bus.emit(
                Message("mycroft.volume.set",
                        {"percent": self.fake_barge_in_volume / 100,  # alsa plugin expects between 0-1
                         "play_sound": False},
                        {"skill_id": "dinkum-listener"})
            )
        self.bus.emit(Message("recognizer_loop:record_begin"))

    def _save_ww(self, audio_bytes, ww_meta, save_path=None):
        if save_path:
            hotword_audio_dir = Path(save_path)
        else:
            hotword_audio_dir = Path(f"{self.default_save_path}/wake_words")
            hotword_audio_dir.mkdir(parents=True, exist_ok=True)

        metafile = self._compile_ww_context(ww_meta["key_phrase"], ww_meta["module"])
        # TODO - do we need to keep this convention? i don't think so...
        #   move to the standard ww_id + timestamp from OPM
        filename = '_'.join(str(metafile[k]) for k in sorted(metafile))

        mic = self.voice_loop.mic
        wav_path = hotword_audio_dir / f"{filename}.wav"
        meta_path = hotword_audio_dir / f"{filename}.json"
        with open(wav_path, "wb") as wav_io, \
                wave.open(wav_io, "wb") as wav_file:
            wav_file.setframerate(mic.sample_rate)
            wav_file.setsampwidth(mic.sample_width)
            wav_file.setnchannels(mic.sample_channels)
            wav_file.writeframes(audio_bytes)
        with open(meta_path, "w") as f:
            json.dump(metafile, f)

        LOG.debug(f"Wrote {wav_path}")
        return f"file://{wav_path.absolute()}"

    def _upload_hotword(self, wav_data, metadata):
        """Upload the wakeword in a background thread."""

        upload_url = Configuration().get("listener", {}).get('wake_word_upload', {}).get('url')

        def upload(wav_data, metadata):
            DatasetApi().upload_wake_word(wav_data,
                                          metadata,
                                          upload_url=upload_url)

        if DatasetApi is not None:
            Thread(target=upload, daemon=True,
                   args=(wav_data, metadata)).start()
        else:
            LOG.debug("`pip install ovos-backend-client` to enable upload")

    @staticmethod
    def _compile_ww_context(key_phrase, ww_module):
        """ creates metadata in the format expected by selene
        while this format is mostly deprecated we want to
        ensure backwards compat and no missing keys"""
        model_hash = '0'
        return {
            'name': key_phrase,
            'engine': md5(ww_module.encode('utf-8')).hexdigest(),
            'time': str(int(1000 * time.time())),
            'sessionId': SessionManager.get().session_id,
            'accountId': "Anon",
            'model': str(model_hash)
        }

    def _hotword_audio(self, audio_bytes: bytes, ww_context: dict):
        """
        Callback method for when a hotword is detected
        @param audio_bytes: Audio that triggered detection
        @param ww_context: Context attached to hotword detection
        """
        payload = ww_context
        context = {'client_name': 'ovos_dinkum_listener',
                   'source': 'audio',  # default native audio source
                   'destination': ["skills"]}
        stt_lang = ww_context.get("stt_lang")  # per wake word lang override in mycroft.conf
        if stt_lang:
            context["lang"] = stt_lang

        try:
            listener = self.config["listener"]
            if listener["record_wake_words"]:
                payload["filename"] = self._save_ww(audio_bytes, ww_context)

            upload_disabled = listener.get('wake_word_upload',
                                           {}).get('disable')
            if self.config['opt_in'] and not upload_disabled:
                self._upload_hotword(audio_bytes, ww_context)

            utterance = ww_context.get("utterance")
            if utterance:
                LOG.debug("Hotword utterance: " + utterance)
                # send the transcribed word on for processing
                payload = {
                    'utterances': [utterance],
                    "lang": stt_lang or Configuration().get("lang", "en-us")
                }
                self.bus.emit(Message("recognizer_loop:utterance",
                                      payload,
                                      context))
                return payload

            # If enabled, play a wave file with a short sound to audibly
            # indicate hotword was detected.
            sound = ww_context.get("sound")
            listen = ww_context.get("listen")
            event = ww_context.get("event")

            if sound:
                LOG.debug(f"Handling listen sound: {sound}")
                audio_context = dict(context)
                audio_context["destination"] = ["audio"]
                self.bus.emit(Message("mycroft.audio.play_sound",
                                      {"uri": sound, "force_unmute": True},
                                      audio_context))
            if listen:
                msg_type = "recognizer_loop:wakeword"
                payload["utterance"] = \
                    ww_context["key_phrase"].replace("_", " ").replace("-", " ")
            elif event:
                msg_type = event
            else:
                if ww_context.get("wakeup"):
                    wordtype = "wakeupword"
                elif ww_context.get("stop"):
                    wordtype = "stopword"
                else:
                    wordtype = "hotword"
                msg_type = f"recognizer_loop:{wordtype}"

            LOG.debug(f"Emitting hotword event: {msg_type}")
            # emit ww event
            self.bus.emit(Message(msg_type, payload, context))

        except Exception:
            LOG.exception("Error while saving STT audio")
        return payload

    def _record_end_signal(self):
        LOG.debug("Record end")
        if self.fake_barge_in:
            LOG.info(f"fake barge-in restoring volume to: {self._default_vol}")
            self.bus.emit(
                Message("mycroft.volume.set",
                        {"percent": self._default_vol / 100,  # alsa plugin expects between 0-1
                         "play_sound": False},
                        {"skill_id": "dinkum-listener"})
            )
        self.bus.emit(Message("recognizer_loop:record_end"))

    def _stt_text(self, transcripts: List[Tuple[str, float]],
                  stt_context: dict):
        # Report utterance to intent service
        if transcripts:
            utts = [u[0] for u in transcripts]  # filter confidence
            lang = stt_context.get("lang") or Configuration().get("lang", "en-us")
            LOG.debug(f"STT: {utts}")
            payload = {"utterances": utts,
                       "lang": lang}
            self.bus.emit(Message("recognizer_loop:utterance", payload, stt_context))
        elif self.voice_loop.listen_mode == ListeningMode.CONTINUOUS:
            LOG.debug("ignoring transcription failure")
        else:
            self.bus.emit(Message("recognizer_loop:speech.recognition.unknown", context=stt_context))

    def _save_stt(self, audio_bytes, stt_meta, save_path=None):
        LOG.info("Saving Utterance Recording")
        if save_path:
            stt_audio_dir = Path(save_path)
        else:
            stt_audio_dir = Path(f"{self.default_save_path}/utterances")
        stt_audio_dir.mkdir(parents=True, exist_ok=True)

        filename = hash_sentence(stt_meta["transcription"])
        mic = self.voice_loop.mic
        wav_path = stt_audio_dir / f"{filename}.wav"
        meta_path = stt_audio_dir / f"{filename}.json"
        with open(wav_path, "wb") as wav_io, \
                wave.open(wav_io, "wb") as wav_file:
            wav_file.setframerate(mic.sample_rate)
            wav_file.setsampwidth(mic.sample_width)
            wav_file.setnchannels(mic.sample_channels)
            wav_file.writeframes(audio_bytes)
        with open(meta_path, "w") as f:
            json.dump(stt_meta, f)

        LOG.debug(f"Wrote {wav_path}")
        return f"file://{wav_path.absolute()}"

    def _upload_stt(self, wav_data, metadata):
        """Upload the STT in a background thread."""

        upload_url = Configuration().get("listener", {}).get('stt_upload', {}).get('url')

        def upload(wav_data, metadata):
            DatasetApi().upload_stt(wav_data, metadata, upload_url=upload_url)

        if DatasetApi:
            Thread(target=upload, daemon=True,
                   args=(wav_data, metadata)).start()
        else:
            LOG.debug("`pip install ovos-backend-client` to enable upload")

    def _stt_audio(self, audio_bytes: bytes, stt_context: dict):
        try:
            listener = self.config["listener"]
            if listener["save_utterances"]:
                stt_context["filename"] = self._save_stt(audio_bytes, stt_context)
                upload_disabled = listener.get('stt_upload', {}).get('disable')
                if self.config['opt_in'] and not upload_disabled:
                    self._upload_stt(audio_bytes, stt_context)
        except Exception:
            LOG.exception("Error while saving STT audio")
        return stt_context

    def _save_recording(self, audio_bytes, stt_meta, save_path=None):
        LOG.info("Saving Recording")
        if save_path:
            rec_audio_dir = Path(save_path)
        else:
            rec_audio_dir = Path(self.default_save_path) / "recordings"
        rec_audio_dir.mkdir(parents=True, exist_ok=True)

        filename = stt_meta.get("recording_name", time.time())
        mic = self.voice_loop.mic
        wav_path = rec_audio_dir / f"{filename}.wav"
        meta_path = rec_audio_dir / f"{filename}.json"
        with open(wav_path, "wb") as wav_io, \
                wave.open(wav_io, "wb") as wav_file:
            wav_file.setframerate(mic.sample_rate)
            wav_file.setsampwidth(mic.sample_width)
            wav_file.setnchannels(mic.sample_channels)
            wav_file.writeframes(audio_bytes)
        with open(meta_path, "w") as f:
            json.dump(stt_meta, f)

        LOG.debug(f"Wrote {wav_path}")
        return f"file://{wav_path.absolute()}"

    def _recording_audio(self, audio_bytes: bytes, stt_context: dict):
        try:
            stt_context["filename"] = self._save_recording(audio_bytes, stt_context)
        except Exception:
            LOG.exception("Error while saving recording audio")
        return stt_context

    # mic bus api
    def _handle_mute(self, message: Message):
        self.voice_loop.is_muted = True

    def _handle_unmute(self, message: Message):
        self.voice_loop.is_muted = False

    def _handle_mute_toggle(self, message: Message):
        self.voice_loop.is_muted = not self.voice_loop.is_muted

    def _handle_listen(self, message: Message):
        if not self._validate_message_context(message) or not self.voice_loop.running:
            # ignore mycroft.mic.listen, it is targeted to an external client
            return
        if self.voice_loop.wake_callback is not None:
            # Emit `recognizer_loop:record_begin`
            self.voice_loop.wake_callback()
        self.voice_loop.reset_speech_timer()
        self.voice_loop.stt_audio_bytes = bytes()
        self.voice_loop.stt.stream_start()
        if self.voice_loop.fallback_stt is not None:
            self.voice_loop.fallback_stt.stream_start()

        if self.config.get('confirm_listening'):
            sound = self.config.get('sounds', {}).get('start_listening')
            if sound:
                self.bus.emit(message.forward("mycroft.audio.play_sound", {"uri": sound}))
                self.voice_loop.state = ListeningState.CONFIRMATION
                try:
                    if sound.startswith("snd/"):
                        dur = get_sound_duration(sound, base_dir=f"{dirname(__file__)}/res")
                    else:
                        dur = get_sound_duration(sound)
                    LOG.debug(f"{sound} duration: {dur} seconds")
                    self.voice_loop.confirmation_seconds_left = dur
                except:
                    self.voice_loop.confirmation_seconds_left = self.voice_loop.confirmation_seconds
        else:
            self.voice_loop.state = ListeningState.BEFORE_COMMAND

    def _handle_mic_get_status(self, message: Message):
        """Query microphone mute status."""
        data = {'muted': self.voice_loop.is_muted}
        self.bus.emit(message.response(data))

    def _handle_audio_start(self, message: Message):
        """audio output started"""
        if self.config.get("listener").get("mute_during_output"):
            self.voice_loop.is_muted = True

    def _handle_audio_end(self, message: Message):
        """audio output ended"""
        if self.config.get("listener").get("mute_during_output"):
            self.voice_loop.is_muted = False  # restore

    def _handle_stop(self, message: Message):
        """Handler for mycroft.stop, i.e. button press."""
        self.voice_loop.is_muted = False  # restore

    # state events
    def _handle_change_state(self, message: Message):
        """Set listening state."""
        # TODO - unify this api, should match ovos-listener exactly
        state = message.data.get("state")
        mode = message.data.get("mode")

        # NOTE: the enums are also strings and will match
        if state:
            if state == ListeningState.SLEEPING:
                self.voice_loop.go_to_sleep()
            elif state == ListeningState.DETECT_WAKEWORD or state == ListeningState.WAITING_CMD:  # "continuous"
                self.voice_loop.reset_state()
            elif state == ListeningState.RECORDING:  # "recording"
                self.voice_loop.start_recording(message.data.get("recording_name"))
            else:
                LOG.error(f"Invalid listening state: {state}")

        if mode:
            if mode == ListeningMode.WAKEWORD:
                self.voice_loop.listen_mode = ListeningMode.WAKEWORD
            elif mode == ListeningMode.CONTINUOUS:
                self.voice_loop.listen_mode = ListeningMode.CONTINUOUS
            elif mode == ListeningMode.HYBRID:
                self.voice_loop.listen_mode = ListeningMode.HYBRID
            elif mode == ListeningMode.SLEEPING:
                self.voice_loop.listen_mode = ListeningMode.SLEEPING
            else:
                LOG.error(f"Invalid listen mode: {mode}")

        self._handle_get_state(message)

    def _handle_get_state(self, message: Message):
        """Query listening state"""
        # TODO - unify this api, should match ovos-listener exactly
        data = {'mode': self.voice_loop.listen_mode,
                "state": self.voice_loop.state}
        self.bus.emit(message.reply("recognizer_loop:state", data))

    def _handle_stop_recording(self, message: Message):
        """Stop current recording session """
        self.voice_loop.stop_recording()
        sound = self.config.get('sounds', {}).get('end_listening')
        if sound:
            self.bus.emit(message.forward("mycroft.audio.play_sound", {"uri": sound}))

    def _handle_extend_listening(self, message: Message):
        """ when a skill is activated (converse) reset the timeout until wakeword is needed again
        only used when in hybrid listening mode """
        if self.voice_loop.listen_mode == ListeningMode.HYBRID:
            self.voice_loop.last_ww = time.time()

    def _handle_sleep(self, message: Message):
        """Put the voice loop to sleep."""
        self.voice_loop.go_to_sleep()

    def _handle_wake_up(self, message: Message):
        """Wake up the voice loop."""
        LOG.debug("SLEEP - wake up triggered from bus event")
        self.voice_loop.wakeup()

    def _handle_sound_played(self, message: Message):
        """Handle response message from audio service."""
        if not self._validate_message_context(message) or not self.voice_loop.running:
            # ignore this sound, it is targeted to an external client
            return
        if self.voice_loop.state == ListeningState.CONFIRMATION:
            self.voice_loop.state = ListeningState.BEFORE_COMMAND

    def _handle_b64_audio(self, message: Message):
        """ transcribe base64 encoded audio """
        b64audio = message.data["audio"]
        lang = message.data.get("lang", self.voice_loop.stt.lang)

        wav_data = base64.b64decode(b64audio)

        audio = bytes2audiodata(wav_data)

        utterances = self.voice_loop.stt.transcribe(audio, lang)
        filtered = [u for u in utterances if u[1] >= self.voice_loop.min_stt_confidence]
        if filtered != utterances:
            LOG.info(f"Ignoring low confidence STT transcriptions: {[u for u in utterances if u not in filtered]}")

        if filtered:
            self.bus.emit(message.forward(
                "recognizer_loop:utterance",
                {"utterances": [u[0] for u in filtered],
                 "lang": lang}))
        else:
            self.bus.emit(message.forward(
                "recognizer_loop:speech.recognition.unknown"))

    # OPM bus api
    def _handle_get_languages_stt(self, message):
        """
        Handle a request for supported STT languages
        :param message: ovos.languages.stt request
        """
        stt_langs = self.voice_loop.stt.available_languages or \
                    [self.config.get('lang') or 'en-us']
        LOG.debug(f"Got stt_langs: {stt_langs}")
        self.bus.emit(message.response({'langs': list(stt_langs)}))

    @staticmethod
    def get_stt_lang_options(lang, blacklist=None):
        blacklist = blacklist or []
        opts = []
        cfgs = get_stt_lang_configs(lang=lang, include_dialects=True)
        for engine, configs in cfgs.items():
            if engine in blacklist:
                continue
            # For Display purposes, we want to show the engine name without the underscore or dash and capitalized all
            plugin_display_name = engine.replace("_", " ").replace("-", " ").title()
            for config in configs:
                config["plugin_name"] = plugin_display_name
                config["engine"] = engine
                config["lang"] = config.get("lang") or lang
                opts.append(config)
        return opts

    @staticmethod
    def get_ww_lang_options(lang, blacklist=None):
        blacklist = blacklist or []
        opts = []
        cfgs = get_ww_lang_configs(lang=lang, include_dialects=True)
        for engine, configs in cfgs.items():
            if engine in blacklist:
                continue
            # For Display purposes, we want to show the engine name without the underscore or dash and capitalized all
            plugin_display_name = engine.replace("_", " ").replace("-", " ").title()
            for config in configs:
                config["plugin_name"] = plugin_display_name
                config["engine"] = engine
                config["lang"] = config.get("lang") or lang
                opts.append(config)
        return opts

    @staticmethod
    def get_vad_options(blacklist=None):
        blacklist = blacklist or []
        tts_opts = []
        cfgs = get_vad_configs()
        for engine, configs in cfgs.items():
            if engine in blacklist:
                continue
            # For Display purposes, we want to show the engine name without the underscore or dash and capitalized all
            plugin_display_name = engine.replace("_", " ").replace("-", " ").title()
            for voice in configs:
                voice["plugin_name"] = plugin_display_name
                voice["engine"] = engine
                tts_opts.append(voice)
        return tts_opts

    def _handle_opm_stt_query(self, message):
        plugs = get_stt_supported_langs()
        configs = {}
        opts = {}
        for lang, m in plugs.items():
            for p in m:
                configs[p] = get_stt_module_configs(p)
            opts[lang] = self.get_stt_lang_options(lang)

        data = {
            "plugins": plugs,
            "langs": list(plugs.keys()),
            "configs": configs,
            "options": opts
        }
        self.bus.emit(message.response(data))

    def _handle_opm_ww_query(self, message):
        plugs = get_ww_supported_langs()
        configs = {}
        opts = {}
        for lang, m in plugs.items():
            for p in m:
                configs[p] = get_ww_module_configs(p)
            opts[lang] = self.get_ww_lang_options(lang)

        data = {
            "plugins": plugs,
            "langs": list(plugs.keys()),
            "configs": configs,
            "options": opts
        }
        self.bus.emit(message.response(data))

    def _handle_opm_vad_query(self, message):
        cfgs = get_vad_configs()
        data = {
            "plugins": list(cfgs.keys()),
            "configs": cfgs,
            "options": self.get_vad_options()
        }
        self.bus.emit(message.response(data))

    def reload_configuration(self):
        """
        Reload configuration and restart loop. Automatically called when
        Configuration object reports a change
        """
        if self._config_hash() == self._applied_config_hash:
            LOG.info(f"No relevant configuration changed")
            return
        LOG.info("Reloading changed configuration")
        if not self._load_lock.acquire(timeout=30):
            raise TimeoutError("Lock not acquired after 30 seconds")
        if self._shutdown_event.is_set():
            LOG.info("Shutting down, skipping config reload")
            self._load_lock.release()
            return
        try:
            LOG.debug("Lock Acquired")
            new_hash = self._config_hash()

            # Configuration changed, update status and reload
            self.status.set_alive()

            if new_hash['stt'] != self._applied_config_hash['stt']:
                LOG.info(f"Reloading STT")
                if self.stt:
                    LOG.debug(f"old={self.stt.__class__}: {self.stt.config}")
                if hasattr(self.stt, "shutdown"):
                    self.stt.shutdown()
                del self.stt
                self.stt = load_stt_module(self.config['stt'])
                self.voice_loop.stt = self.stt
                if self.stt:
                    LOG.debug(f"new={self.stt.__class__}: {self.stt.config}")

            if new_hash['fallback'] != self._applied_config_hash['fallback']:
                LOG.info(f"Reloading Fallback STT")
                if self.fallback_stt:
                    LOG.debug(f"old={self.fallback_stt.__class__}: "
                              f"{self.fallback_stt.config}")
                if hasattr(self.fallback_stt, "shutdown"):
                    self.fallback_stt.shutdown()
                del self.fallback_stt
                self.fallback_stt = load_fallback_stt(self.config['stt'])
                self.voice_loop.fallback_stt = self.fallback_stt
                if self.fallback_stt:
                    LOG.debug(f"new={self.fallback_stt.__class__}: "
                              f"{self.fallback_stt.config}")

            if new_hash['hotwords'] != self._applied_config_hash['hotwords']:
                LOG.info(f"Reloading Hotwords")
                LOG.debug(f"old={self.hotwords.applied_hotwords_config}")
                self._reload_event.clear()
                self.voice_loop.stop()
                self.hotwords.shutdown()
                self.hotwords.load_hotword_engines()
                LOG.debug(f"new={self.hotwords.applied_hotwords_config}")

            if new_hash['loop'] != self._applied_config_hash['loop']:
                LOG.info(f"Reloading Listener")
                self._reload_event.clear()
                self.voice_loop.stop()

                if hasattr(self.vad, "stop"):
                    self.vad.stop()
                del self.vad
                self.vad = OVOSVADFactory.create(self.config)

                self.mic.stop()
                del self.mic
                microphone_config = self.config.get("microphone", {})
                microphone_config.setdefault('module',
                                             'ovos-microphone-plugin-alsa')
                self.mic = OVOSMicrophoneFactory.create(microphone_config)
                self.mic.start()

                # Update voice_loop with new parameters
                self.voice_loop.mic = self.mic
                self.voice_loop.vad = self.vad
                listener_config = self.config['listener']
                self.voice_loop.speech_seconds = \
                    listener_config.get("speech_begin", 0.3)
                self.voice_loop.silence_seconds = \
                    listener_config.get("silence_end", 0.7)
                self.voice_loop.timeout_seconds = listener_config.get(
                    "recording_timeout", 10)
                self.voice_loop.num_stt_rewind_chunks = listener_config.get(
                    "utterance_chunks_to_rewind", 2)
                self.voice_loop.num_hotword_keep_chunks = listener_config.get(
                    "wakeword_chunks_to_save", 15)
            if not self.voice_loop.running:
                self.voice_loop.start()
                self._reload_event.set()

            self._applied_config_hash = new_hash
            self.status.set_ready()
            LOG.info("Reload Completed")
        except Exception as e:
            LOG.exception(e)
            self.status.set_error(e)
        finally:
            self._load_lock.release()
